#include "php_orbit.h"
#include "phorb.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define CORBA_dup(t, v)	(t)CORBA_Object_duplicate((CORBA_Object)(v), NULL)

static zend_class_entry *corbatc_class;

void phorb_init_typecode_class(zend_class_entry *tcclass)
{
	corbatc_class = tcclass;
}

void phorb_fini_typecode_class(void)
{
}

zval *phorb_init_typecode(zval *res)
{
	object_init_ex(res, corbatc_class);

	return res;
}

ulong phorb_register_typecode(CORBA_TypeCode tc)
{
	return phorb_register_object((CORBA_Object)tc);
}

CORBA_TypeCode phorb_find_typecode(phorbIDL *idl, const char *repoid)
{
	GSList *list;
	CORBA_TypeCode ret = CORBA_OBJECT_NIL;
	OrbSLS_FETCH();

	if (idl)
		ret = g_hash_table_lookup(idl->typecodes, repoid);
	else {
		for(list = OrbSG(idls); list; list = list->next) {
			idl = (phorbIDL *)list->data;
			ret = g_hash_table_lookup(idl->typecodes, repoid);
			if (ret)
				break;
		}
	}

	return ret ? CORBA_dup(CORBA_TypeCode, ret) : CORBA_OBJECT_NIL;
}

void phorb_store_typecode(phorbIDL *idl, CORBA_TypeCode tc)
{
	tc = CORBA_dup(CORBA_TypeCode, tc);
	g_hash_table_insert(idl->typecodes, tc->repo_id, tc);
}

#define ADD_BASIC(type)	phorb_store_typecode(idl, TC_CORBA_##type)

void phorb_setup_typecodes(phorbIDL *idl)
{
	ADD_BASIC(short);
	ADD_BASIC(long);
	ADD_BASIC(long_long);
	ADD_BASIC(unsigned_short);
	ADD_BASIC(unsigned_long);
	ADD_BASIC(unsigned_long_long);
	ADD_BASIC(float);
	ADD_BASIC(double);
	ADD_BASIC(long_double);
	ADD_BASIC(boolean);
	ADD_BASIC(char);
	ADD_BASIC(wchar);
	ADD_BASIC(octet);
	ADD_BASIC(any);
	ADD_BASIC(TypeCode);
	ADD_BASIC(Object);
	ADD_BASIC(string);
	ADD_BASIC(wstring);
}

#undef ADD_BASIC

void phorb_invoke_typecode_method(INTERNAL_FUNCTION_PARAMETERS,
                                  zend_property_reference *props)
{
	zval *handle, *pobj = props->object;
	zend_overloaded_element	*fname =
		(zend_overloaded_element *)props->elements_list->tail->data;
	char *method = Z_STRVAL(fname->element);
	char *name;
	int name_len;
	CORBA_TypeCode tc;
	OrbSLS_FETCH();

	if (zend_llist_count(props->elements_list) != 1
	    || strcmp(method, "corba_typecode") != 0) {
		zend_error(E_WARNING, "Can't invoke methods on typecodes");

		RETVAL_NULL();
		goto leave;
	}

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s",
	                          &name, &name_len) == FAILURE)
		goto error;

	g_assert(Z_TYPE_P(pobj) == IS_OBJECT);

	tc = phorb_find_typecode(NULL, name);
	if (!tc) {
		zend_error(E_WARNING,
		           "Could not find TypeCode for %s in "
		           "CORBA_TypeCode()", name);
		goto error;
	}

	MAKE_STD_ZVAL(handle);
	ZVAL_RESOURCE(handle, phorb_register_typecode(tc));
	zend_hash_index_update(Z_OBJPROP_P(pobj), 0, &handle,
		               sizeof(zval *), NULL);

leave:
	zval_dtor(&fname->element);
	return;

error:
	zval_dtor(pobj);
	ZVAL_NULL(pobj);

	goto leave;
}

zval phorb_get_typecode_attribute(zend_property_reference *props)
{
	zval dres;
	zval *res = &dres;
	zval **handle, *pobj = props->object;
	CORBA_TypeCode tc;
	zend_overloaded_element *aname = 
		(zend_overloaded_element *)props->elements_list->tail->data;
	char *attribute = Z_STRVAL(aname->element);
	int type;

	INIT_ZVAL(dres);

	zend_hash_index_find(Z_OBJPROP_P(pobj), 0, (void *)&handle);
	tc = (CORBA_TypeCode)zend_list_find(Z_RESVAL_PP(handle), &type);

	if (strcmp(attribute, "kind") == 0) {
		ZVAL_STRING(res, TC_CORBA_TCKind->subnames[tc->kind], 1);
	} else if (strcmp(attribute, "length") == 0) {
		ZVAL_LONG(res, tc->length);
	} else if (strcmp(attribute, "sub_parts") == 0) {
		ZVAL_LONG(res, tc->sub_parts);
	} else if (strcmp(attribute, "name") == 0) {
		ZVAL_STRING(res, tc->name, 1);
	} else if (strcmp(attribute, "repo_id") == 0) {
		ZVAL_STRING(res, tc->repo_id, 1);
	} else {
		zend_error(E_WARNING, "Non-existing attribute `%s'",
		           attribute);
	}

	zval_dtor(&aname->element);
	return dres;
}

int phorb_set_typecode_attribute(zend_property_reference *props, zval *value)
{
	zend_overloaded_element *aname = 
		(zend_overloaded_element *)props->elements_list->tail->data;

	zend_error(E_WARNING, "Typecode attributes are readonly");

	zval_dtor(&aname->element);
	return FAILURE;
}

void phorb_serialize_typecode_sleep(INTERNAL_FUNCTION_PARAMETERS)
{
	CORBA_TypeCode tc;
	zval **handle;
	char *repoid;
	int type;

	zend_hash_index_find(Z_OBJPROP_P(this_ptr), 0, (void *)&handle);
	tc = (CORBA_TypeCode)zend_list_find(Z_RESVAL_PP(handle), &type);
	array_init(return_value);
	repoid = (char *)tc->repo_id;

	if (!repoid || !repoid[0])
		repoid = "<unknown>";

	add_property_string(this_ptr, "_repoid", repoid, 1);
	add_next_index_string(return_value, "_repoid", 1);
}

void phorb_serialize_typecode_wakeup(INTERNAL_FUNCTION_PARAMETERS)
{
	CORBA_TypeCode tc;
	zval *handle, **repoid;
	int type;

	zend_hash_find(Z_OBJPROP_P(this_ptr), "_repoid", 8, (void *)&repoid);
	if (!repoid) {
		zend_error(E_ERROR, "Bad serialization of CORBA_TypeCode "
		                    "object, no repoid found");
		RETURN_FALSE;
	}
	if (Z_TYPE_PP(repoid) != IS_STRING) {
		zend_error(E_ERROR, "Bad serialization of CORBA_TypeCode "
		                    "object, repoid not a string");
		RETURN_FALSE;
	}

	tc = phorb_find_typecode(NULL, Z_STRVAL_PP(repoid));
	if (!tc) {
		zend_error(E_WARNING, "Could not find repoid `%s' in "
		                      "typecodes", Z_STRVAL_PP(repoid));
		RETURN_FALSE;
	}
	zval_ptr_dtor(repoid);
	zend_hash_del(Z_OBJPROP_P(this_ptr), "_repoid", 8);

	MAKE_STD_ZVAL(handle);
	ZVAL_RESOURCE(handle, phorb_register_typecode(tc));

	zend_hash_index_update(Z_OBJPROP_P(this_ptr), 0, &handle,
	                       sizeof(zval *), NULL);

	RETURN_TRUE;
}

CORBA_long phorb_enum_find_member(CORBA_TypeCode tc, zval *val)
{
	char *str;
	CORBA_long i;

	convert_to_string(val);
	str = Z_STRVAL_P(val);
	if (!str)
		return -1;

	for(i = 0; i < tc->sub_parts; i++)
		if (strcmp(tc->subnames[i], str) == 0)
			return i;

	return -1;
}

#define ZVAL_RES_FIND_MEMBER(kind, type) \
	case CORBA_tk_##kind: do { \
		CORBA_##type val = Z_RESVAL_P(discriminator); \
		\
		for(i = 0; i < tc->sub_parts; i++) { \
			CORBA_##type label = \
				*(CORBA_##type *)tc->sublabels[i]; \
			\
			if (label == val) \
				return i; \
		} \
		break; \
	} while(0)

CORBA_long phorb_union_find_arm(CORBA_TypeCode tc, zval *discriminator)
{
	CORBA_unsigned_long i;

	/* assuming discriminator IS_RESOURCE */

	switch(tc->discriminator->kind) {
	ZVAL_RES_FIND_MEMBER(char, char);
	ZVAL_RES_FIND_MEMBER(octet, octet);
	ZVAL_RES_FIND_MEMBER(short, short);
	ZVAL_RES_FIND_MEMBER(ushort, unsigned_short);
	ZVAL_RES_FIND_MEMBER(long, long);
	ZVAL_RES_FIND_MEMBER(ulong, unsigned_long);
	ZVAL_RES_FIND_MEMBER(longlong, long_long);
	ZVAL_RES_FIND_MEMBER(ulonglong, unsigned_long_long);
	case CORBA_tk_enum: {
		CORBA_unsigned_long val =
			phorb_enum_find_member(
				tc->discriminator, discriminator);

		for(i = 0; i < tc->sub_parts; i++)
			if (tc->sublabels[i] == val)
				return i;
		break;
	}
	case CORBA_tk_boolean: {
		CORBA_boolean val = Z_RESVAL_P(discriminator);

		for(i = 0; i < tc->sub_parts; i++)
			if (!tc->sublabels[i] == !val)
				return i;
		break;
	}
	default:
		zend_error(E_WARNING,
		           "Unsupported discriminator type %d",
		           tc->discriminator->kind);
	}

	return (tc->default_index >= 0) ? tc->default_index : -1;
}

#undef ZVAL_RES_FIND_MEMBER
